package CD;

import MDLUtilities.Marquee1In_1Out;
import CD.AssocSchema.atype;
import PrologDB.DB;
import PrologDB.Table;
import PrologDB.Tuple;
import java.io.FileNotFoundException;

/**
 * unparses a cd specification specification into a plantuml spec
 */
public class CD2PlantUML extends MDLUtilities.RunningBear {

    static final String q = "\"";
    static final String virtual = "#lime";

    private static Table tab, attr, assoc, sub;

    /**
     * main prograqm
     *
     * @param args -- args
     */
    public static void main(String... args) {

        // Step 1: standard marquee processing
        Marquee1In_1Out mark = new Marquee1In_1Out(CD2PlantUML.class, ".cd.pl", ".class.plantuml", args);
        RBSetup(mark, args);

        cd2plantuml(db);
    }

    public static void cd2plantuml(DB db, String outputFileName) throws FileNotFoundException {
        openOut(outputFileName);
        cd2plantuml(db);
    }

    public static void cd2plantuml(DB db) {
        // step 1: open tables
        tab = db.getTableEH("tab");
        attr = db.getTableEH("attr");
        assoc = db.getTableEH("assoc");
        sub = db.getTableEH("sub");

        // Step 2: write header
        l("@startuml");
        l();
        l("note as N1");
        l("    %s    ",db.getName());
        l("end note");
        l();

        // Step 3: output each table
        for (Tuple t : tab.tuples()) {

            String ab = t.getBool("abs") ? "abstract " : "";
            if (t.getBool("virt")==false) {
                l("%sclass %s {", ab, t.get("name"));
            } else {
                l("%sclass %s %s{", ab, t.get("name"),virtual);
            }
            for (Tuple a : t.rightSemiJoin("id", attr, "attrof").tuples()) {
                String aname = a.get("name");
                if (aname.equals("id")) {
                    continue;
                }
//                boolean inherited = suptFields.anyMatch(sf -> sf.is("name", aname));
//                if (inherited) {
//                    continue;
//                }
                String name2display = a.get("name");
                switch (a.get("ptype")) {
                    case "int":
                    case "bool":
                    case "float":
                    case "dbl":
                        name2display = name2display + " : " + a.get("ptype");
                        break;
                    case "qstr":
                        name2display = q + name2display + q;
                        break;
                    case "str":
                    default:

                }
                if (a.getBool("virt")==false){
                   l("  %s", name2display);
                } else {
                   l("  +%s",name2display);
                }
            }
            l("}");
            l();
        }

        // Step 4: output each inheritance
        for (Tuple s : sub.tuples()) {
            boolean supIsV = tab.follow(s.get("suptab")).getBool("virt");
            boolean subIsV = tab.followEH(s.get("subtab")).getBool("virt"); 
            
            String supname = tab.followEH(s.get("suptab")).get("name");
            String subname = tab.followEH(s.get("subtab")).get("name");
            if (subIsV || supIsV) {
                l("%s <|-- %s %s", supname, subname,virtual);
            } else {
                l("%s <|-- %s", supname, subname);
            }
        }

        // Step 4: output each association
        for (Tuple a : assoc.tuples()) {
            String left, right, connector;
            String class1name = tab.followEH(a.get("class1")).get("name");
            String class2name = tab.followEH(a.get("class2")).get("name");
            String role1 = a.get("role1");
            String role2 = a.get("role2");
            atype at = atype.convert(a.get("atype1"));
            atype at2 = atype.convert(a.get("atype2"));
            switch (at) {
                case DIAMOND:
                    left = role1;
                    right = role2(role2,at2,a);
                    connector = "o--";
                    break;
                case BLACK_DIAMOND:
                    left = role1;
                    right = role2(role2,at2,a);//role2 + card(a, '2');
                    connector = "*--";
                    break;
                case BLANK:
                    left = role1;
                    right = role2(role2,at2,a);//role2;
                    connector = "--";
                    break;
                case ONESTAR:
                    left = role1 + " 1+";
                    right = role2(role2,at2,a);//role2;
                    connector = "--";
                    break;
                case CUSTOM:
                    left = role1 + card(a, '1');
                    right = role2(role2,at2,a);//role2 + card(a, '2');
                    connector = "--";
                    break;
                default:
                    throw new RuntimeException("SHOULD NEVER GET HERE!");
            }
            if (a.getBool("virt")== false) {
               l("%s \"%s\" %s \"%s\" %s", class1name, left, connector, right, class2name);
            } else {
               l("%s \"%s\" %s \"%s\" %s %s", class1name, left, connector, right, class2name,virtual);
            }
        }

        // Step X: finish
        l("@enduml");
        closeOut();
    }

    static String role2(String name, atype at, Tuple a) {
        switch (at) {
            case DIAMOND:
                return name + " <>";
            case BLACK_DIAMOND:
                return name + " 1";
            case BLANK:
                return name;
            case ONESTAR:
                return name + " 1+";
            case CUSTOM:
                return name + card(a,'2');
            default:
                throw new RuntimeException("SHOULD NEVER GET HERE!");
        }
    }

    static String card(Tuple a, char whichone) {
        String min = a.get("min" + whichone);
        String max = a.get("max" + whichone);
        if (min.equals("0") && max.equals("-1")) {
            return "";
        }
        if (min.equals("1") && max.equals("-1")) {
            return " 1+";
        }
        if (min.equals("1") && max.equals("1")) {
            return " 1";
        }
        if (min.equals("0") && max.equals("1")) {
            return " 0:1";
        }
        return " " + min + ":" + max;
    }

    static Tuple findSuperTable(Tuple tb) {
        String tbid = tb.getId();
        for (Tuple t : sub.tuples()) {
            if (t.is("subtab", tbid)) {
                return tab.followEH(t.get("suptab"));
            }
        }
        return null;
    }
}
