package CD;

import MDLUtilities.Marquee1In_1Out;
import Meta4.Main;
import Parsing.TBPrims.SaveToken.Opt;
import PrologDB.DB;
import PrologDB.DBSchema;
import PrologDB.Table;
import PrologDB.TableSchema;

public class CD2Schema {
    static String strEnd = ":str";
    static String schemaEnd = ".schema.pl";

    static String outputFileName, cdspecName, appName;

    static Table tab, attr, assoc, sub;

    public static void main(String... args) {
        // Step 0 : standard marquee processing
        Marquee1In_1Out mark = new Marquee1In_1Out(CD2Yuml.class, ".cd.pl", schemaEnd, args);
        cdspecName = mark.getInputFileName();
        outputFileName = mark.getOutputFileName();
        appName = mark.getAppName();

        // Step 1: read in the input database
        DB db = DB.read(cdspecName);
        main(db, appName, outputFileName);
    }
    
    public static void main(DB db, String appName, String outputFileName) {
        
    }
    
    public static void addAssocFieldsAndTables(CDSpec s) {
        // WARNING -- this method modifies CDSchema s
        // it must be the last (or before finishAndPrint) method that is invoked!!!
        s.setUnfinished();
        
        // Step 1: review the associations, add fields to existing tables
        //         and add association tables
        for (AssocSchema as : s.getAssocSchemas()) {
            if (as.isVirtual()) {
                continue;
            }
            TableSchema ts;
            
            // Step 1a: getAssocSchemaSpec raw data
            String e1 = as.getClass1();
            String e2 = as.getClass2();
            String r1 = as.getRole1();
            String r2 = as.getRole2();
            String ml = as.getName();
            
            // Step 1b: deal with nulls
            r1 = r1.equals("null")?ml:r1;
            r2 = r2.equals("null")?ml:r2;
            
            AssocSchema.atype at = as.getAt1();
            
            switch(at) {
                case BLACK_DIAMOND:
                    ts = s.getTableSchemaEH(e2);
                    ts.add(mkCol(r1,e1,false));
                    break;
                case DIAMOND:
                    ts = s.getTableSchemaEH(e2);
                    ts.add(mkCol(r1,e1,true));
                    break;
                case BLANK:
                    ts = new TableSchema(ml);
                    ts.add(mkCol(r1,e1,false),mkCol(r2,e2,false));
                    s.add(ts);
                    break;
            }
        }
        // done!
        s.addIdentifiersToAllNonSubtables();
        s.setFinished();
    }

    public static void finishAndPrint(CDSpec s) {
        // Step 1: a CD schema is also an ooschema.  propagate its attributes
        //         and produce a printable db
        DBSchema schema = s.finishAndPropagateAttributes();
        String fname = String.format("%s/%s/%s.schema.pl",Main.AoclDir, s.getName(),schema.getName());
        schema.print(fname);
    }
    
    private static String mkCol(String fieldName, String kind, boolean optional) {
        return fieldName + ":" + (optional?(Opt.tok+" "):"") + kind;
    }
    
//    // this method modifies DB db!! call at the very end!
//    public static void main(DB db, String appName, String outputFileName) {
//        tab = db.getTableEH("tab");
//        attr = db.getTableEH("attr");
//        assoc = db.getTableEH("assoc");
//        sub = db.getTableEH("sub");
//
//        // will modify db inplace to add fields for associations
//        // Step 2: precondition -- every association must be a DIAMOND or BLACK_DIAMOND
//        for (Tuple a : assoc.tuples()) {
//            if (a.getBool("virt")) 
//                continue;
//            // perform add fields ONLY if association is real
//         
//            if (a.is("atype1", "BLACK_DIAMOND") || a.is("atype1", "DIAMOND")) {
//                continue;
//            }
//            // manufacture a new association-class table if association is not virtual
//            if (a.getSchema().isVirtual())
//                continue;
//            
//            String name = a.get("name");
//            boolean duplicate = tab.select(t->t.is("name",name)).isNotEmpty();
//            if (duplicate) {
//                System.err.format("cannot create table for association (%s) as a table with that name already exists",name);
//                return;
//            }
//            String classleft = a.get("class1");
//            String classrite = a.get("class2");
//            String roleleft  = a.get("role1");
//            String rolerite  = a.get("role2");
//            tab.add(name,name,"false","false");  // add non-abstract table
//            attr.add(name+"_"+classleft,roleleft,classleft,name,"false");
//            attr.add(name+"_"+classrite,rolerite,classrite,name,"false");
//        }
//
//        // Step 3: make (empty) tables only for non-virtuals
//        OOSchema dbs = new OOSchema(appName);
//        for (Tuple t : tab.tuples()) {
//            if (t.getSchema().isVirtual())
//                continue;
//            // Step 3a: create empty table schema
//            TableSchema ts = new TableSchema(t.get("name"));
//            if (t.is("abs")) {
//                ts.makeAbstract();
//            }
//            dbs.add(ts);
//            String tid = t.getId();
//
//            // Step 3b: now add attributes
//            Column c;
//            for (Tuple a : t.rightSemiJoin("id", attr, "attrof").tuples()) {
//                //String name, opts opt, String kind, boolean isQuoted
//                String ptype = a.get("ptype");
//                boolean isq = ptype.equals("qstr");
//                c = new Column(a.get("name"), opts.empty, ptype, isq);
//                ts.add(c);
//            }
//
//            // Step 3c: now add attributes of associations
//            for (Tuple lnk : assoc.tuples()) {
//                if (lnk.getBool("virt")==true) {  // ignore virtual associations
//                    continue;
//                }
//                if (lnk.is("class2", tid)) {
//                    String c1name = getTableName(lnk.get("class1"));
//                    if (lnk.is("atype1", "BLACK_DIAMOND")) {
//                        c = new Column(lnk.get("role1"), opts.empty, c1name, false);
//                        ts.add(c);
//                    } else
//                    if (lnk.is("atype1","DIAMOND")) {
//                        c = new Column(lnk.get("role1"), opts.option, c1name, false);
//                        ts.add(c);
//                    }
//                }
//            }
//        }
//
//        // Step 4: add subtables
//        SubTableSchema sts;
//        TableSchema ts;
//        for (Tuple s : sub.tuples()) {
//            String supName = getTableName(s.get("suptab"));
//            Table supTable = db.getTableEH(supName);
//            if (supTable.getTableSchema().isVirtual())
//                continue;
//            sts = dbs.getSubTableSchema(supName);
//            if (sts == null) {
//                ts = dbs.getTableSchemaEH(supName);
//                sts = new SubTableSchema(ts);
//                dbs.add(sts);
//            }
//            String subName = getTableName(s.get("subtab"));
//            TableSchema sub = dbs.getTableSchemaEH(subName);
//            sts.add(sub);
//        }
//
//        // Step 5: flatten and write write to disk
//        
//        dbs.addIdentifiersToAllNonSubtables();
//        String schemaName = MDLUtilities.Utils.getAppName(outputFileName, schemaEnd);
//        dbs.rename(schemaName);
//        DBSchema dbf = dbs.finishAndPropagateAttributes();
//        dbf.print(outputFileName);
//    }
//
//    static String getTableName(String id) {
//        return tab.followEH(id).get("name");
//    }

}
