/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package CD;

import MDLUtilities.Marquee4Conform;
import CD.AssocSchema.atype;
import PrologDB.DB;
import PrologDB.ErrorReport;
import PrologDB.Table;
import PrologDB.Tuple;
import Violett.ClassConform;

/**
 *
 * @author don
 */
public class CDConform {

    public static void main(String... args) {
        // Step 1: Standard marquee processing
        Marquee4Conform mark = new Marquee4Conform(ClassConform.class, ".cd.pl", args);
        String inputFileName = mark.getInputFileName();

        // Step 2: read database and create error handler
        DB db = DB.read(inputFileName);
        check(db);
    }

    public static void check(DB db) {
        Table tab = db.getTableEH("tab");
        Table attr = db.getTableEH("attr");
        Table sub = db.getTableEH("sub");
        Table assoc = db.getTableEH("assoc");
        // there exists a cons table but it is not yet in use
        ErrorReport er = new ErrorReport();

        db.StandardConform(er, false);
        // does all these tests
        //        tab.isUnique("id", er);
        //        attr.isUnique("id", er);
        //        sub.isUnique("id",er);
        //        assoc.isUnique("id",er);
        //find the first attribute of each table; it must be 'id'
        //        attr.isLegit("attrof", tab, "id", er);
        //        assoc.isLegit("class1",tab, "id", er);
        //        assoc.isLegit("class2",tab, "id", er);
        //        sub.isLegit("suptab",tab, "id", er);
         //       sub.isLegit("subtab", tab,"id",er);

        // all names are unique
        tab.isUnique("name", er);
        assoc.isUnique("name", er);

        // take the union of attributes and possible association fields
        // they must have unique names
        for (Tuple t : tab.tuples()) {
            Table an = t.rightSemiJoin("id", attr, "attrof").project("name");
            String tid = t.getId();
            for (Tuple a : assoc.tuples()) {
                if (a.is("class1",tid) && a.get("atype1").endsWith("DIAMOND")) {
                    an.add(a.get("role1"));
                }
//                if (a.is("class1", tid)) {
//                    an.add(a.get("role1"));
//                }
//                if (a.is("class2", tid)) {
//                    an.add(a.get("role2"));
//                }
            }
            an.duplicates().error(er, "class %s may have @2+ attributes with name %s",
                    s -> t.get("name"), s -> s.get("name"));
        }

        // atypes for associations should be legal
        for (Tuple t : assoc.tuples()) {
            atype a1 = atype.convert(t.get("atype1"));
            atype a2 = atype.convert(t.get("atype2"));
            
            int min1 = t.getInt("min1");
            int min2 = t.getInt("min2");
            int max1 = t.getInt("max1");
            int max2 = t.getInt("max2");
            
            if (a1 == atype.BLACK_DIAMOND ||a1 == atype.DIAMOND) {
                if (min1!=a1.min || max1!=a1.getMax()) 
                    er.add("assoc(%s,...) atype1 has wrong min(%d) or max(%d)",t.getId(),min1,max1);
            };
            
            if (a2 == atype.BLACK_DIAMOND ||a2 == atype.DIAMOND) {
                if (min2!=a2.min || max2!=a2.getMax()) 
                    er.add("assoc(%s,...) atype2 has wrong min(%d) or max(%d)",t.getId(),min2,max2);
            };

//            if (a1 == atype.UNKNOWN) {
//                er.add("association id=%s has unknown atype %s", t.getId(), t.get("atype1"));
//            }
            if (0 > min1 ) {
                er.add("association id=%s has a negative minimum", t.getId());
            }
            if ( max2!=-1 && 0>max2) {
                er.add("association id=%s has a negative maximum", t.getId());
            }

            if (!(min1 <= convert(max1) && min2 <= convert(max2))) {
                er.add("association id=%s has min>max", t.getId());
            }
            
            if (/*er.hasErrors() &&*/ ! a1.geq(a2)) { // must reverse ends
                er.add("assoc(%s...) reverse association ends! Swap end (%s,%s) with (%s,%s)!",t.getId(),
                        t.get("class1"),t.get("role1"),t.get("class2"),t.get("role2"));
//                String c1 = t.get("class1");
//                String c2 = t.get("class2");
//                String r1 = t.get("role1");
//                String r2 = t.get("role2");
//                t.setValues(t.getId(),t.get("name"),c2,r2,a2.toString(),min2+"",max2+"",
//                                                    c1,r1,a1.toString(),min1+"",max1+"");
            }
        }

        // End report all errors
        er.printEH();
    }
    
    private static int convert(int i) { return (i==-1)?atype.infinity:i; }

}
