package Common;

import java.io.PrintStream;
import java.util.LinkedList;
import java.util.function.Consumer;
import java.util.function.Predicate;

public abstract class GTab<Database, T extends Id<Database>> {
    
    protected Database db;
    
    public LinkedList<T> tuples = new LinkedList<>();
    
    public void initTupleList(T t) {
        tuples = new LinkedList<>();
        if (t != null)
            tuples.add(t);
    }
    
    public abstract String mastHead();
    
    public void setDB(T t) {
        t.setDB(db);
    }
    
    public void add(T d) {
        tuples.add(d);
        setDB(d);
    }
    
    public T getTuple(String id) {
        for (T t : tuples) {
            if (t.getId().equals(id)) {
                return t;
            }
        }
        return null;
    }

    public T getTupleEH(String id) {
        T result = getTuple(id);
        if (result != null) {
            return result;
        }
        throw new RuntimeException("no tuple found in " + this.getClass().getName() + " with id = " + id);
    }
    
    protected <This extends GTab<Database,T>> This select(Predicate<T> p, This shorterTable) {
        // Step 1: dup over every tuple left t that satisfies p
        tuples.stream().filter(p).forEach(tt -> shorterTable.add(tt));

        // Step 2: output row-reduced tbl
        return shorterTable;
    }
    
    public void forEach(Consumer<? super T> action) {
        tuples.forEach(action);
    }
    
    public void print(PrintStream ps) {
        ps.print(mastHead());
        tuples.forEach(t->ps.print(t));
        ps.println();
    }

    public void print() {
        print(System.out);
    }
}
