package M4;

import CD.AssocSchema;
import CD.CDConform;
import CD.CDSpec;
import MDLUtilities.Marquee1In_1Out;
import Parsing.TBPrims.SaveToken.Opt;
import PrologDB.DB;
import PrologDB.DBSchema;
import PrologDB.TableSchema;


// translates a cdspec file into mdelite schema
// calling sequence <x.cdspec.pl> <x.schema.pl>
public class CDSpec2Schema  {
    public static String schemaEnd = ".schema.pl";
    public static String tmp = "tmp.cd.pl";

    public static void main(String... args)  {
        try {
        // Step 1: standard marquee processing
        Marquee1In_1Out mark = new Marquee1In_1Out(CDSpec2Schema.class,CDSpec.cdspecEnd, schemaEnd, args);
        String inputFileName = mark.getInputFileName();
        String outputFileName = mark.getOutputFileName();
        String appName = mark.getAppName();
        
        // Step 2: straightline code
        CDSpec cd = CDSpec.read(inputFileName);
        DB cddb = cd.toCD();
        CDConform.check(cddb);

        addAssocFieldsAndTables(cd);
        finishAndPrint(cd,outputFileName);
        } catch (Exception ex) {
            throw new RuntimeException(ex);
        }
    }
    
        public static void addAssocFieldsAndTables(CDSpec s) {
        // WARNING -- this method modifies CDSchema s
        // it must be the last (or before finishAndPrint) method that is invoked!!!
        s.setUnfinished();
        
        // Step 1: review the associations, add fields to existing tables
        //         and add association tables
        for (AssocSchema as : s.getAssocSchemas()) {
            if (as.isVirtual()) {
                continue;
            }
            TableSchema ts;
            
            // Step 1a: getAssocSchemaSpec raw data
            String e1 = as.getClass1();
            String e2 = as.getClass2();
            String r1 = as.getRole1();
            String r2 = as.getRole2();
            String ml = as.getName();
            
            // Step 1b: deal with nulls
            r1 = r1.equals("null")?ml:r1;
            r2 = r2.equals("null")?ml:r2;
            
            AssocSchema.atype at = as.getAt1();
            
            switch(at) {
                case BLACK_DIAMOND:
                    ts = s.getTableSchemaEH(e2);
                    ts.add(mkCol(r1,e1,false));
                    break;
                case DIAMOND:
                    ts = s.getTableSchemaEH(e2);
                    ts.add(mkCol(r1,e1,true));
                    break;
                case BLANK:
                    ts = new TableSchema(ml);
                    ts.add(mkCol(r1,e1,false),mkCol(r2,e2,false));
                    s.add(ts);
                    break;
            }
        }
        // done!
        s.addIdentifiersToAllNonSubtables();
        s.setFinished();
    }

    public static void finishAndPrint(CDSpec s, String outputFileName) {
        // Step 1: a CD schema is also an ooschema.  propagate its attributes
        //         and produce a printable db
        DBSchema db = s.finishAndPropagateAttributes();
        db.print(outputFileName);
    }
    
    private static String mkCol(String fieldName, String kind, boolean optional) {
        return fieldName + ":" + (optional?(Opt.tok+" "):"") + kind;
    }
}

