package M4Utilities;

import Parsing.DBPrimitives.FixedStrTokens.DBase;
import Parsing.DBPrimitives.FixedStrTokens.OODB;

/** known parsing errors for MDELite */
public enum ParsE implements ErrInt<ParsE> {

    csvTableNameMismatch("CVS tuple table name TABLENAME1 does not match CSV file name TABLENAME2"),
    dbDeclMissing("no dbase/oodb/classDiagram declaration in FILENAME"),
    dbDeclsTooMany("more than one dbase/oodb/classDiagram declaration in FILENAME"),
    dbInappropriate(DBase.tok+ " declaration is inappropriate for an "+OODB.tok +" schema"),
    ooInappropriate(OODB.tok + " declaration is inappropriate for an "+DBase.tok+" schema"),
    ioerror("In file FILENAME : ERROR"),
    parseError("parse error MESSAGE"),
    parseIllformedNumber("ill-formed number starting with STRING"),
    patternEmpty("pattern is empty"),
    runAwayEntry("unended yuml box entry STR"),
    runAwayString("unended quoted string STR"),
    streamEndReached("end of line reached: TOKEN expected"),
    subTableDefBeforeDB("subtable defined before oodb or dbase"),
    tableDefBeforeDB("table defined before oodb or dbase"),
    tableOnly1Allowed("only one table definition permitted in FILENAME"),
    tableSchemaParseError("unable to parse table in FILENAME \n ERROR"),
    tableUndefined("in file FILENAME : tuple declaration for undefined table TABLENAME"),
    tupleHasInsufficientValues("in file FILENAME : insufficient number of values for a tuple of table TABLENAME"),
    tupleHasInsufficientValues2("insufficient number of values for a tuple of table TABLENAME"),
    yumlDuplicateBoxDecl("duplicate declaration for class NAME"),
    yumlIntHasFields("yuml interface NAME has fields"),
    yumlLineParser("yuml line parser not initialized"),
    yumlNoteHasFields("yuml note NAME has fields"),
    yumlNoteHasMethods("yuml note NAME has methods"),
    
    noTaBaDeclParsed("no taba decl parsed yet"), ;

    private final String msg;

    /** java required constructor */
    ParsE(String msg) {
        this.msg = msg;
    }

    /** get parse error message/template */
    public String getMsg() {
        return msg;
    }
    
    /**
     * @return the enums of ParsE in an array
     */
    @Override
    public ErrInt<ParsE>[] vals() {
        return ParsE.values();
    }

    public static String makeString(ParsE error, int lineno, Object... args) {
        String front = (lineno < 0) ? "" : "line " + lineno + ": ";
        String back = Utils.makeString(error.msg, args);
        return front + back;
    }

    /** method that is used to create a RuntimeException to throw
     * 
     * @param error -- which parse error
     * @param lineno -- at what line number (-1 if unknown(
     * @param args -- arguments of parse error message
     * @return RuntimeException object to throw
     */
    public static RuntimeException toss(ParsE error, int lineno, Object... args) {
        if (error.msg.contains("FILENAME")) { args[0] = Utils.shortFileName((String) args[0]); }
        String msg = makeString(error, lineno, args);
        return new RuntimeException(msg);
    }
}
