package Meta4;

import CD.CDSpec;
import static MDLUtilities.RunningBear.l;
import static Meta4.PrintersAndUtilities.makeStackOfSuperSchemas;
import PrologDB.Column;
import PrologDB.OOSchema;
import PrologDB.SubTableSchema;
import PrologDB.TableSchema;
import java.util.Stack;

public class DatabaseFile extends MDLUtilities.RunningBear {

    static OOSchema db;

    public static void makeDatabaseFile(CDSpec s) {
        db = s;
        String filename = Main.AoclDir + "/" + s.getName() + "/Database.java";
        openOut(filename);
        l("package %s.%s;", Main.AoclDir, s.getName());
        l();
        l("import PrologDB.DB;");
        l("import PrologDB.Table;");
        l("import PrologDB.ErrorReport;");
        l("import java.io.PrintStream;");
        l();
        l("public class Database {");
        l();
        apply("    public %sTable %s = %sTable.BaseRelation(this);", false);
        l();
        l("    private DB db;");
        l();
        l("    public Database() {}");
        l();
        l("    public Database(String filename) {");
        l("        // Step 1: Read the database, throw exception if errors found and halt");
        l("        db = DB.read(filename);");
        l();
        l("        // Step 2: access MDELite tables");
        apply("        Table _%s = db.getTableEH(\"%s\");", true);
        l();
        l("        // Step 3: pass1 -- load primitive data of all non-abstract + non-virtual tables");
        apply("        _%s.tuplesLocal().forEach(t->%s.makeAndAdd(t));", true);
        l();
        l("        // Step 4: pass2 -- now replace MDELite tuple ids with Java object ids");
        l("        ErrorReport er = new ErrorReport();");
        apply("        %s.tuplesLocal().forEach(t->t.finishTuple(this,er));", true);
        l();
        l("        // Step 5: pass3 -- EXPERIMENTAL init of virtual tables --- must manually fill in expressions");
        aVirt("        %sPopulate(this);",s);
        l();
        l("        er.printEH(System.out);");
        l("    }");
//        VTableMethods();
        printers(s);
        VPopulateMethods(s);
        l("}");
        closeOut();
    }

    public static void apply(String pattern, boolean checkAV) {
        for (TableSchema ts : db.getTableSchemas()) {
            if (checkAV) {
                if (ts.isVirtual() || ts.isAbstract()) {
                    continue;
                }
            }
            String tsname = ts.getName();
            l(pattern, tsname, tsname, tsname, tsname, tsname, tsname);
        }
    }

    public static void aVirt(String pattern, CDSpec s) {
        for (String x : s.getExtras()) {
            l(pattern,x,x,x);
        }
    }
    
    public static void VPopulateMethods(CDSpec s) {
        for (String x : s.getExtras()) {
            AppendFile(x+".txt");
        }
    }
        
//        // initialize virtual tables first
//        for (TableSchema ts : db.getTableSchemas()) {
//            if (ts.isVirtual()) {
//                String tsname = ts.getName();
//                l(pattern, pkname, tsname, tsname, tsname, tsname, tsname, tsname);
//            }
//        }
//        l();
//        l("        // complete concrete tables that have virtual fields");
//        l();
//
//        // initialize partially virtual (tables with virtual fields) next
//        for (TableSchema ts : db.getTableSchemas()) {
//            if (!ts.isVirtual() && hasVirtualColumns(ts, db)) {
//                String tsname = ts.getName();
//                l(pattern, pkname, tsname, tsname, tsname, tsname, tsname, tsname);
//            }
//        }
//    }

    public static boolean hasVirtualColumns(TableSchema ts, OOSchema s) {
        Stack<TableSchema> stack = makeStackOfSuperSchemas(ts, s);
        while (!stack.isEmpty()) {
            ts = stack.pop();
            for (Column c : ts.getUnInheritedColumns()) {
                if (c.isVirtual()) {
                    return true;
                }
            }
        }
        return false;
    }

    public static void VTableMethods() {

        boolean first = true;
        for (TableSchema ts : db.getTableSchemas()) {
            boolean hasV = ts.isVirtual() || ts.hasVirtualFields();
            if (hasV) {
                if (first) {
                    l(1);
                    l("    // ******START OF methods to populate virtual tables -- experimental*****");
                    first = false;
                }
                String tsn = ts.getName();
                l();
                l("    public void %sPopulate() { /** FILL IN **/ }", tsn);
            }
        }
        if (!first) {
            l(2);
            l("    // ******END OF methods to populate virtual tables -- experimental*****");
            l(2);
        }
    }

    public static void printers(OOSchema s) {
        l();
        l("    public void print() { print(System.out); }");
        l();
        l("    public void prettyprint() { prettyprint(System.out); }");
        l();
        pprint(s, true);
        pprint(s, false);
    }

    private static void pprint(OOSchema s, boolean prettyprint) {
        String prefix = prettyprint ? "pretty" : "";
        l("    public void %sprint(PrintStream out) {", prefix);
        p("        out.format(\"dbase(%s,[", s.getName());
        String comma = "";
        for (TableSchema ts : db.getTableSchemas()) {
            String tsname = ts.getName();
            p("%s%s", comma, tsname);
            comma = ",";
        }
        l("]).\\n\\n\");");
        for (TableSchema ts : db.getTableSchemas()) {
            String tsname = ts.getName();
            l("        %s.%sprintLocal(out);", tsname, prefix);
        }
        for (SubTableSchema sts : db.getSubTableSchemas()) {
            p("        out.format(\"subtable(%s,[", sts.getSuper().getName());
            comma = "";
            for (TableSchema ts : sts.getSubTableSchemas()) {
                p("%s%s", comma, ts.getName());
                comma = ",";
            }
            l("]).\\n\");");
        }
        l("    }");
        l();
    }
}
