package Meta4;

import CD.CDConform;
import CD.CDSpec;
import MDLUtilities.Marquee1In_1DirOut;
import MDLUtilities.Utils;
import PrologDB.DB;
import java.io.File;
import java.io.IOException;

public class Main {
    
    public static final String AoclDir = "Aocl";
    static final String tmpfile =  "tmp.class.plantuml";

    public static void main(String... args) throws IOException {

        // Step 1: standard marquee processing
        Marquee1In_1DirOut mark = new Marquee1In_1DirOut(Main.class, CDSpec.cdspecEnd,AoclDir, args);
        String infile = mark.getInputFileName();
        String AppName = AoclDir + "/"+ mark.getAppName();
        
        // Step 2: make sure that the AoclDir is ok
        File od = new File(AoclDir);
        if (od.exists()) {
            if (!od.isDirectory()) {
                System.out.format("%s is a file not a directory - cannot proceed\n", AoclDir);
                System.exit(0);
            }
        } else {
            if (!od.mkdir()) {
                System.out.format("cannot create directory %s -- stop\n", AoclDir);
                System.exit(0);
            }
        }

        // Step 3: prepare the directory
        try {
            Utils.deleteDirectory(AppName);
        } catch (Exception e) {
            System.out.println("cannot delete directory "+ AppName);
            System.exit(0);
        }
        if (!new File(AppName).mkdir()) {
            throw new RuntimeException("Could not create " + AppName + " directory");
        }

        // Step 4: read the schema into a useful form
        CDSpec s = CDSpec.read(infile);  
        DB db = s.toCD();       
        CDConform.check(db);
        db.print("tmp.cd.pl");
        
        // Step 5: create png file
        CD.CD2PlantUML.cd2plantuml(db, tmpfile);
        M4.PlantUML2Png.main(tmpfile, AppName+"/"+mark.getAppName()+".class.png");

        // Step 6: produce the files
        LangFile.makeLangFile(s);
//        CDSpec2Schema.addAssocFieldsAndTables(s);// this method modifies s
        SchemaFile.addAssocFieldsAndTables(s); // this method modifies s
        DatabaseFile.makeDatabaseFile(s);
        TupleFiles.makeTupleFiles(s);
        TableFiles.makeTableFiles(s);
//        CDSpec2Schema.finishAndPrint(s, schemaFile);// also modifies s
        SchemaFile.finishAndPrint(s);    // also modifies s
        
        System.out.println("output is in directory "+ AppName);
    }
    
    static void debug(CDSpec s, String dif) {
        System.out.format("---- start %s ----\n",dif);
        s.print();
        System.out.format("---- end   %s ----\n",dif);
    }

}

