package Meta4;

import MDLUtilities.RunningBear;
import static MDLUtilities.RunningBear.l;
import static MDLUtilities.RunningBear.p;
import Parsing.TBPrims.SaveToken.Opt;
import PrologDB.Column;
import PrologDB.OOSchema;
import PrologDB.PrimType;
import PrologDB.SubTableSchema;
import PrologDB.TableSchema;
import java.util.LinkedList;
import java.util.List;
import java.util.Stack;

public class PrintersAndUtilities extends RunningBear {

    public static void makeListOfSubSchemas(LinkedList<TableSchema> result, TableSchema ts, OOSchema s) {
        SubTableSchema sts = s.getSubTableSchema(ts);
        if (sts == null) {
            return;
        }
        List<TableSchema> list = sts.getSubTableSchemas();
        if (list == null) {
            return;
        }
        for (TableSchema tsch : list) {
            result.add(tsch);
            makeListOfSubSchemas(result, tsch, s);
        }
    }

    public static Stack<TableSchema> makeStackOfSuperSchemas(TableSchema ts, OOSchema s) {
        Stack<TableSchema> result = new Stack<>();
        result.push(ts);
        while ((ts = getSuperSchema(ts, s)) != null && !ts.isVirtual()) {
            result.push(ts);
        }
        return result;
    }

    public static TableSchema getSuperSchema(TableSchema ts, OOSchema s) {
        for (SubTableSchema sts : s.getSubTableSchemas()) {
            // find tableschema that is the parent of ts
            if (sts.hasAsASubschema(ts) && !sts.getSuper().isVirtual()) {
                ts = sts.getSuper();
                return ts;
            }
        }
        return null;
    }

    public static boolean hasNonPrimitiveVirtualColumns(TableSchema ts, OOSchema s) {
        Stack<TableSchema> stack = makeStackOfSuperSchemas(ts, s);
        while (!stack.isEmpty()) {
            ts = stack.pop();
            for (Column c : ts.getUnInheritedColumns()) {
                String okind = c.getKind();
                if (!PrimType.isPrim(okind) && !c.isVirtual()) {
                    return true;
                }
            }
        }
        return false;
    }

    public static void displayFieldsOO(OOSchema s, TableSchema ts, Printer printer) {
        Stack<TableSchema> stack;
        TableSchema sup = ts.getSuperTableOf();
        if (sup == null || sup.isVirtual()) {
            stack = new Stack<>();
            stack.push(ts);
        } else {
            stack = makeStackOfSuperSchemas(ts, s);
        }
        while (!stack.isEmpty()) {
            ts = stack.pop();
            PrintersAndUtilities.displayFields(ts, printer);
        }
    }

    public static void displayFields(TableSchema ts, Printer pr) {
        for (Column c : ts.getUnInheritedColumns()) {
            String cname = c.getName();
            boolean isq = c.isQuoted();
            Column.opts opt = c.getOpt();
            String okind = c.getKind();
            boolean isPrim = PrimType.isPrim(okind);
            boolean v = c.isVirtual();

            if (isPrim) {
                PrimType p = PrimType.convert2PrimType(okind);

                // Step 1b: crank out java definition for attributes
                switch (opt) {
                    case empty:
                    case option:
                        pr.printRealColumn(isq, p.javaName(), cname, c.isVirtual());
                    default:
                }
            } else { // pointers using foreign keys
                switch (opt) {
                    case empty:
                    case option:
//                        pr.printRealColumn(isq, okind, cname);
                        pr.printFakeColumn(isq, okind, cname, v);
                    // fall through
                    case table:
                }
            }
            pr.commaUpdate();
        }
    }

    public static void displayAllFields(TableSchema ts, OOSchema s, Printer pr) {
        Stack<TableSchema> stack = makeStackOfSuperSchemas(ts, s);
        while (!stack.isEmpty()) {
            TableSchema TS = stack.pop();
            for (Column c : TS.getColumns()) {
                String cname = c.getName();
                boolean isq = c.isQuoted();
                Column.opts opt = c.getOpt();
                String okind = c.getKind();
                boolean isPrim = PrimType.isPrim(okind);
                boolean v = c.isVirtual();

                if (isPrim) {
                    PrimType p = PrimType.convert2PrimType(okind);

                    // Step 1b: crank out java definition for attributes
                    switch (opt) {
                        case empty:
                        case option:
                            pr.printRealColumn(isq, p.javaName(), cname, v);
                        default:
                    }
                } else { // pointers using foreign keys
                    switch (opt) {
                        case empty:
                        case option:
//                        pr.printRealColumn(isq, okind, cname);
                            pr.printFakeColumn(isq, okind, cname, v);
                        case table:
                    }
                }
                pr.commaUpdate();
            }
        }
    }

    /*--------------------printer utilities-----------------*/
    public static interface Printer {

        void printRealColumn(boolean isQ, String kind, String cname, boolean isV);

        void printFakeColumn(boolean isQ, String kind, String cname, boolean isV);

        void commaUpdate();
    }

    static class fieldPrinter implements Printer {

        String comma = "";

        @Override
        public void commaUpdate() {
            comma = "";
        }

        @Override
        public void printRealColumn(boolean isQ, String kind, String cname, boolean isV) {
            if (!cname.equals("id")) {
                String comment = isV ? "// vfield" : "";
                l("    public %s %s;  %s", kind.replace(Opt.tok + " ", ""), cname, comment);
            }
        }

        @Override
        public void printFakeColumn(boolean isQ, String kind, String cname, boolean isV) {
            String comment = isV ? "   // vfield" : "";
            l("    public %s %s;%s", kind.replace(Opt.tok + " ", ""), cname, comment);
            if (isV) {
                return;
            }
            l("    protected String _%s;  // needed by database loader", cname);
        }
    }

    static class constPrinter implements Printer {

        String comma = "";

        @Override
        public void commaUpdate() {
            comma = ", ";
        }

        @Override
        public void printRealColumn(boolean isQ, String kind, String cname, boolean isV) {
            if (isV) {
                return;
            }
            if (PrimType.isExtendedPrim(kind)) {
                p("%s%s %s", comma, kind, cname);
            }
        }

        @Override
        public void printFakeColumn(boolean isQ, String kind, String cname, boolean isV) {
        }
    }

    static class allConstArgsPrinter implements Printer {

        String comma = "";

        @Override
        public void commaUpdate() {
            comma = ", ";
        }

        @Override
        public void printRealColumn(boolean isQ, String kind, String cname, boolean isV) {
            if (isV) {
                return;
            }
            p("%s%s %s", comma, kind.replace(Opt.tok + " ", ""), cname);
        }

        @Override
        public void printFakeColumn(boolean isQ, String kind, String cname, boolean isV) {
            if (isV) {
                return;
            }
            p("%s%s %s", comma, kind.replace(Opt.tok + " ", ""), cname);
        }
    }

    static class initAllFieldsPrinter implements Printer {

        String comma = "";

        @Override
        public void commaUpdate() {
            comma = "";
        }

        @Override
        public void printRealColumn(boolean isQ, String kind, String cname, boolean isV) {
            if (isV) {
                return;
            }
            l("       this.%s = %s;", cname, cname);
        }

        @Override
        public void printFakeColumn(boolean isQ, String kind, String cname, boolean isV) {
            if (isV) {
                l("       this.%s = null;", cname);
            } else {
                l("       this.%s = %s;", cname, cname);
            }
        }
    }

    static class printPrinter implements Printer {

        String comma = "";

        @Override
        public void commaUpdate() {
            comma = ",";
        }

        @Override
        public void printRealColumn(boolean isQ, String kind, String cname, boolean isV) {
            if (isV) {
                return;
            }
            p("%sa.%s", comma, cname);
        }

        @Override
        public void printFakeColumn(boolean isQ, String kind, String cname, boolean isV) {
            p("%s__(a.%s)", comma, cname);
        }
    }

    static class initFieldsForConstructorPrinter implements Printer {

        String comma = "";

        @Override
        public void commaUpdate() {
            comma = "";
        }

        @Override
        public void printRealColumn(boolean isQ, String kind, String cname, boolean isV) {
            if (isV) {
                return;
            }
//            if (PrimType.isExtendedPrim(kind)) {
            l("       this.%s = %s;", cname, cname);
//            } else {
//                l("       this.%s = null;", cname);
//            }
        }

        @Override
        public void printFakeColumn(boolean isQ, String kind, String cname, boolean isV) {
            l("       this.%s = null;", cname);
        }
    }

    static class initFieldsPrinter implements Printer {

        String comma = "";

        @Override
        public void commaUpdate() {
            comma = "";
        }

        @Override
        public void printRealColumn(boolean isQ, String kind, String cname, boolean isV) {
            if (isV) {
                return;
            }
            switch (kind) {
                case "String":
                    l("       %s = t.get(\"%s\");", cname, cname);
                    break;
                case "boolean":
                    l("       %s = t.getBool(\"%s\");", cname, cname);
                    break;
                case "int":
                    l("       %s = t.getInt(\"%s\");", cname, cname);
                    break;
                case "float":
                    l("       %s = t.getFloat(\"%s\");", cname, cname);
                    break;
                case "double":
                    l("       %s = t.getDouble(\"%s\");", cname, cname);
                    break;
                default:
                    break;
            }
        }

        @Override
        public void printFakeColumn(boolean isQ, String kind, String cname, boolean isV) {
            if (isV) {
                l("       %s = null;", cname);
            } else {
                l("       _%s = t.get(\"%s\");", cname, cname);
            }
        }
    }

    static class make1Printer implements Printer {

        String comma = "";

        @Override
        public void commaUpdate() {
            comma = ", ";
        }

        @Override
        public void printRealColumn(boolean isQ, String kind, String cname, boolean isV) {
            if (isV) {
                return;
            }
            switch (kind) {
                case "String":
                    p("%st.get(\"%s\")", comma, cname);
                    break;
                case "boolean":
                    p("%st.getBool(\"%s\")", comma, cname);
                    break;
                case "int":
                    p("%st.getInt(\"%s\")", comma, cname);
                    break;
                case "float":
                    p("%st.getFloat(\"%s\")", comma, cname);
                    break;
                default:
                    break;
            }
        }

        @Override
        public void printFakeColumn(boolean isQ, String kind, String cname, boolean isV) {
        }
    }

    static class make2Printer implements Printer {

        @Override
        public void commaUpdate() {
        }

        @Override
        public void printRealColumn(boolean isQ, String kind, String cname, boolean isV) {
        }

        @Override
        public void printFakeColumn(boolean isQ, String kind, String cname, boolean isV) {
            l("       w._%s = t.get(\"%s\");", cname, cname);
        }
    }

    static class finishPrinter implements Printer {

        @Override
        public void commaUpdate() {
        }

        @Override
        public void printRealColumn(boolean isQ, String kind, String cname, boolean isV) {
        }

        @Override
        public void printFakeColumn(boolean isQ, String kind, String cname, boolean isV) {
            if (isV) {
                return;
            }
            l("       if(!_%s.equals(\"null\"))", cname);
            l("           %s = db.%s.getTuple(_%s,er);", cname, kind.replace(Opt.tok + " ", ""), cname);
        }
    }

    static class gettersPrinter implements Printer {

        @Override
        public void commaUpdate() {
        }

        @Override
        public void printRealColumn(boolean isQ, String kind, String cname, boolean isV) {
//            if (isV) {
//                return;
//            }
//            if (PrimType.isPrim(kind)) {
//                l();
//                l("    public %s %s() {", kind, cname);
//                l("       return %s;", cname);
//                l("    }");
//            }
        }

        @Override
        public void printFakeColumn(boolean isQ, String kind, String cname, boolean isV) {
        }
    }

    static class getIdPrinter implements Printer {

        @Override
        public void commaUpdate() {
        }

        @Override
        public void printRealColumn(boolean isQ, String kind, String cname, boolean isV) {
            l();
            l("    @Override");
            l("    public %s getId() {", kind);
            l("       return %s;", cname);
            l("    }");
        }

        @Override
        public void printFakeColumn(boolean isQ, String kind, String cname, boolean isV) {
        }
    }

    static class toString1Printer implements Printer {

        String comma = "";

        @Override
        public void commaUpdate() {
            comma = ",";
        }

        @Override
        public void printRealColumn(boolean isQ, String kind, String cname, boolean isV) {
            if (isV) {
                return;
            }
            if (isQ) {
                p("%s'%s'", comma, "%s");
            } else {
                p("%s%s", comma, "%s");
            }
        }

        @Override
        public void printFakeColumn(boolean isQ, String kind, String cname, boolean isV) {
        }
    }

    static class namesPrinter implements Printer {

        String comma = "";

        @Override
        public void commaUpdate() {
            comma = ",";
        }

        @Override
        public void printRealColumn(boolean isQ, String kind, String cname, boolean isV) {
            if (isV) {
                return;
            }
            if (PrimType.QStr.javaName().equals(kind) && isQ) {
                p("%sq(%s)", comma, cname);
            } else {
                p("%s%s", comma, cname);
            }
        }

        @Override
        public void printFakeColumn(boolean isQ, String kind, String cname, boolean isV) {
            if (isV) {
                return;
            }
            p("%s__(%s)", comma, cname);
        }
    }

    static class make3Printer implements PrintersAndUtilities.Printer {

        String comma = "";

        @Override
        public void commaUpdate() {
            comma = ",";
        }

        @Override
        public void printRealColumn(boolean isQ, String kind, String cname, boolean isV) {
            if (isV) {
                return;
            }
            switch (kind) {
                case "int":
                    p("%st.getInt(\"%s\")", comma, cname);
                    return;
                case "boolean":
                    p("%st.getBool(\"%s\")", comma, cname);
                    return;
                case "float":
                    p("%st.getFloat(\"%s\")", comma, cname);
                    return;
                case "double":
                    p("%st.getDouble(\"%s\")", comma, cname);
                    return;
                case "String":
                    p("%st.get(\"%s\")", comma, cname);
                default:
            }
        }

        @Override
        public void printFakeColumn(boolean isQ, String kind, String cname, boolean isV) {
        }
    }

    static class make4Printer implements PrintersAndUtilities.Printer {

        @Override
        public void commaUpdate() {
        }

        @Override
        public void printRealColumn(boolean isQ, String kind, String cname, boolean isV) {
        }

        @Override
        public void printFakeColumn(boolean isQ, String kind, String cname, boolean isV) {
            if (isV) {
                return;
            }
            l("       r._%s = t.get(\"%s\");", cname, cname);
        }
    }

    static class formatPrinter implements PrintersAndUtilities.Printer {

        String comma = "";

        @Override
        public void commaUpdate() {
            comma = ",";
        }

        @Override
        public void printRealColumn(boolean isQ, String kind, String cname, boolean isV) {
            if (isV) {
                return;
            }
            switch (kind) {
                case "int":
                    p("%s%s", comma, "%d");
                    return;
                case "boolean":
                    p("%s%s", comma, "%b");
                    return;
                case "float":
                    p("%s%s", comma, "%f");
                    return;
                case "double":
                    p("%s%s", comma, "%f");
                    return;
                default:
                case "String":
                    p("%s%s", comma, "%s");
                    return;
            }
        }

        @Override
        public void printFakeColumn(boolean isQ, String kind, String cname, boolean isV) {
            if (isV) {
                return;
            }
            p("%s%s", comma, "%s");
        }
    }

    static class make0Printer implements PrintersAndUtilities.Printer {

        @Override
        public void commaUpdate() {
        }

        @Override
        public void printRealColumn(boolean isQ, String kind, String cname, boolean isV) {
            if (isV) {
                return;
            }
            p(",t.%s", cname);
        }

        @Override
        public void printFakeColumn(boolean isQ, String kind, String cname, boolean isV) {
        }
    }

    static class ColumnNamePrinter implements PrintersAndUtilities.Printer {

        String comma = "";

        @Override
        public void commaUpdate() {
            comma = ",";
        }

        @Override
        public void printRealColumn(boolean isQ, String kind, String cname, boolean isV) {
            if (isV) {
                return;
            }
            String kkind = kind;
            switch (kind) {
                case "String":
                    kkind = "string";
                    break;
                default:
            }
            if (isQ) {
                p("%s'%s'", comma, cname);
                return;
            }
            if (kkind.equals("string")) {
                p("%s%s", comma, cname);
                return;
            }
            p("%s%s:%s", comma, cname, kkind);
        }

        @Override
        public void printFakeColumn(boolean isQ, String kind, String cname, boolean isV) {
            p("%s%s:%s", comma, cname, kind);
        }
    }

    static class ColumnValuePrinter implements PrintersAndUtilities.Printer {

        @Override
        public void commaUpdate() {
        }

        @Override
        public void printRealColumn(boolean isQ, String kind, String cname, boolean isV) {
            if (isV) {
                return;
            }
            if (PrimType.isExtendedPrim(kind)) {
                p(",t.%s", cname);
            } else {
                p(",__(t.%s)", cname);
            }
        }

        @Override
        public void printFakeColumn(boolean isQ, String kind, String cname, boolean isV) {
        }
    }
}
