package Meta4;

import CD.CDSpec;
import static MDLUtilities.RunningBear.l;
import static MDLUtilities.RunningBear.p;
import PrologDB.TableSchema;
import static Meta4.PrintersAndUtilities.*;
import CD.AssocSchema;
import CD.AssocSchema.atype;

public class TupleFiles extends MDLUtilities.RunningBear {

    public static void makeTupleFiles(CDSpec s) {
        String suprName;

        String AppName = s.getName();
        for (TableSchema ts : s.getTableSchemas()) {
            String tabName = ts.getName();
            TableSchema supr = getSuperSchema(ts, s);
            if (supr == null || supr.isVirtual()) {
                suprName = null;
            } else {
                suprName = supr.getName();
            }
            String fileName = String.format("%s/%s/%s.java", Main.AoclDir, AppName, tabName);
            openOut(fileName);

            header(tabName, AppName, suprName);
            fields(ts);
            constructor(ts, s);
            initFieldAndMakeMethods(ts, s);
            getters(ts, s);
            toStringMethod(ts, s);
            displaySemiJoinMethods(ts, s);
            displayToAssocMethod(ts, s);
            finishTupleFile(tabName);
            closeOut();
        }
    }

    // supName == null 
    public static void header(String tupName, String AppName, String supName) {
        l("package %s.%s;", Main.AoclDir, AppName);
        l();
        l("import PrologDB.Tuple;");
        l("import PrologDB.ErrorReport;");
        l("import java.util.function.Function;");

        if (supName == null) {
            l("import Meta4Framework.TuPle;");
            l();
            l("public class %s extends TuPle<Database> {", tupName);
        } else {
            l();
            l("public class %s extends %s {", tupName, supName);
        }

        l("    @Override");
        l("    public String TupleTypeName() { return \"%s\"; }", tupName);
        l();
        l("    public <TABLE> TABLE apply%s(Function<%s,TABLE> f) { return f.apply(this); }",tupName,tupName);
        l();
    }

    public static void fields(TableSchema ts) {
        displayFields(ts, new fieldPrinter());
        if (ts.isVirtual()) {
            l("    public Object _src; // reference real object");
        } 
        if (ts.size() != 0) {
            l();
        }

    }

    public static void constructor(TableSchema ts, CDSpec s) {
        p("    public %s(", ts.getName());
        displayFieldsOO(s, ts, new allConstArgsPrinter());
        if (ts.isVirtual()) {
            p(", Object _src");
        }
        l(") {");
        displayFieldsOO(s, ts, new initAllFieldsPrinter());
        if (ts.isVirtual()) {
            l("       this._src = _src;");
        }
        l("    }");
        l();
        l("    // constructors used in database loading");
        l("    protected %s() {}", ts.getName());
        l();
        if (hasNonPrimitiveVirtualColumns(ts, s)) {
            p("    public %s(", ts.getName());
            displayFieldsOO(s, ts, new constPrinter());
            l(") {");
            displayFieldsOO(s, ts, new initFieldsForConstructorPrinter());
            l("    }");
            l();
        }
    }

    public static void initFieldAndMakeMethods(TableSchema ts, CDSpec s) {
        String name = ts.getName();
        l("    private %s initFields(Tuple t) {", name);
        displayFieldsOO(s, ts, new initFieldsPrinter());
        l("       return this;");
        l("    }");
        l();
        l("    public static %s make(Tuple t) {", name);
        l("       return new %s().initFields(t);", name);
        l("    }");
        l();
        l("    @Override");
        l("    public void finishTuple(Database db, ErrorReport er) {");
        displayFieldsOO(s, ts, new finishPrinter());
        l("    }");
        l("    // end of methods used for loading a database");
        l();
    }

    public static void getters(TableSchema ts, CDSpec s) {
        displayFields(ts, new gettersPrinter());
    }

    public static void toStringMethod(TableSchema ts, CDSpec s) {
        p("    static final String fmt = \"%s(", ts.getName());
        displayFieldsOO(s, ts, new PrintersAndUtilities.formatPrinter());
        l(").\\n\";");
        l();
        l("    @Override");
        l("    public String toString() {");
        p("       return String.format(fmt,");
        displayFieldsOO(s, ts, new namesPrinter());
        l(");");
        l("    }");
        l();
        l("    @Override");
        l("    public String toString(String fmt) {");
        p("       return String.format(fmt,");
        displayFieldsOO(s, ts, new namesPrinter());
        l(");");
        l("    }");
        l();
    }

    public static void displaySemiJoinMethods(TableSchema ts, CDSpec s) {
        String tname = ts.getName();
        for (AssocSchema as : s.getAssocSchemas()) {
            // Step 1 : unpack and see if relevant
            String e1 = as.getClass1();
            String e2 = as.getClass2();

            if ((s.isGEQ(e1, tname) || s.isGEQ(e2, tname)) == false) {
                continue; // this assocschema is not relevant
            }
            String r1 = as.getRole1();
            String r2 = as.getRole2();
            String ml = as.getName();

            // Step 2: deal with nulls
            r1 = r1.equals("null") ? ml : r1;
            r2 = r2.equals("null") ? ml : r2;

            atype at = as.getAt1();
            switch (at) {
                case BLACK_DIAMOND:
                    if (s.isGEQ(e1, tname)) { //if (tname.equals(e1)) {
                        diamondRight(e2, r2, r1);
                    //} else {
                    }
                    if (s.isGEQ(e2, tname)) {
                        diamondLeft(e1, r1, r2);
                    }
                    break;
                case DIAMOND:
                    if (s.isGEQ(e1, tname)) { //if (tname.equals(e1)) {
                        diamondRight(e2, r2, r1);
                    //} else {
                    }
                    if (s.isGEQ(e2, tname)) {
                        diamondLeft(e1, r1, r2);
                    }
                    break;
                case BLANK:
                    if (s.isGEQ(e1, tname)) { //if (tname.equals(e1)) {
                        nmleft(e2, r2, ml, r1);
                    }
                    if (s.isGEQ(e2, tname)) {
                    //if (tname.equals(e2)) {
                        nmleft(e1, r1, ml, r2);
                    }
                    break;
            }
        }
    }
        
    public static void displayToAssocMethod(TableSchema ts, CDSpec s) {
        String tname = ts.getName();
        for (AssocSchema as : s.getAssocSchemas()) {
            // Step 1 : unpack and see if relevant
            String e1 = as.getClass1();
            String e2 = as.getClass2();

            String r1 = as.getRole1();
            String r2 = as.getRole2();
            String ml = as.getName();

            // Step 2: deal with nulls
            r1 = r1.equals("null") ? ml : r1;
            r2 = r2.equals("null") ? ml : r2;

            atype at = as.getAt1();
            switch (at) {
                case BLACK_DIAMOND:
                    if (tname.equals(e2)) {
                        l("    public %s %s_() {", e1, r1);
                        l("       return this.%s;", r1);
                        l("    }  // diamondleft");
                        l();
                    }
                    break;
                
                default:
            }
        }
    }

    private static void nmleft(String e2, String r2, String ml, String r1) {
        l("    public %sTable %s() {", e2, r2);
        l("       %sTable result = new %sTable(db);", e2, e2);
        l("       for(%s r : db.%s.tuples()) {", ml, ml);
        l("          if(r.%s==this && !result.contains(r.%s)) {", r1, r2);
        l("             result.add(r.%s);", r2);
        l("          }");
        l("       }");
        l("       return result; // nmleft");
        l("    }");
        l();
    }

    private static void diamondLeft(String e1, String r1, String r2) {
        l("    public %sTable %s() {", e1, r1);
        l("       return new %sTable(db,this.%s);", e1, r1);
        l("    }  // diamondleft");
        l();
    }

    private static void diamondRight(String e2, String r2, String r1) {
//        l("    public %sTable %s() {", e2, r2);
//        l("       for(%s a : db.%s.tuples()) {", e2, e2);
//        l("          if(a.%s==this) {", r1);
//        l("             return new %sTable(db,a.%s);",e2,r2);
//        l("          }");
//        l("       }");
//        l("       throw new RuntimeException(\"db inconsistency -- object doesn't exist\");");
//        l("    }  // diamondRight");
//        l();
        l("    public %sTable %s() {", e2, r2);
        l("       %sTable result = new %sTable(db);", e2, e2);
        l("       for(%s a : db.%s.tuples()) {", e2, e2);
        l("          if(a.%s==this) {", r1);
        l("             result.add(a);", r1);
        l("          }");
        l("       }");
        l("       return result; ///diamondRight");
        l("   }");
        l();
    }
    
    private static void finishTupleFile(String tabName) {
        l("    public %sTable to%sTable() { return new %sTable(db,this); }",tabName,tabName, tabName);
        l("}");
    }
}
