package Meta4Framework;

import Meta4Framework.A.Reducer;
import PrologDB.ErrorReport;
import PrologDB.Tuple;
import java.io.PrintStream;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.stream.Stream;
import Meta4Framework.O.*;

public abstract class getKey<Database, T extends TuPle<Database>, TBL extends getKey<Database, T, TBL>> extends CoMMoN<Database> {

    public abstract String TupleTypeName();

    protected boolean computedTable = true;

    protected LinkedList<T> tuples = new LinkedList<>();

    public LinkedList<T> tuplesLocal() {
        return tuples;
    }

    public abstract LinkedList<T> tuples();

    public void initTupleList(T t) {
        tuples = new LinkedList<>();
        if (t != null) {
            tuples.add(t);
        }
    }

    // tuple insertion utilities
    private boolean addPreface(T t) {
        if (t == null) {
            return true;
        }
        t.setDB(db);
        return false;
    }

    public T add(T t) {
        if (addPreface(t)) {
            return null;
        }
        tuples.add(t);
        return t;
    }

    public T addFirst(T t) {
        if (addPreface(t)) {
            return null;
        }
        tuples.addFirst(t);
        return t;
    }

    public T addLast(T t) {
        if (addPreface(t)) {
            return null;
        }
        tuples.addLast(t);
        return t;
    }

    // tuple deletion utilities
    public abstract boolean deleteFromSubTables(T t);

    public abstract void cascadingDeleteLocal(T t);

    public boolean delete(T t) {
        return tuples.remove(t) || deleteFromSubTables(t);
    }

    public boolean cascadingDelete(T t) {
        boolean result = tuples.remove(t) || deleteFromSubTables(t);
        if (result) {
            cascadingDeleteLocal(t);
        }
        return result;
    }

    public void deleteEH(T t) {
        if (delete(t)) {
            return;
        }
        String e = String.format("tuple of type %s and id=%s does not exist and can't be deleted", TupleTypeName(), t.id);
        throw new RuntimeException(e);
    }

    public void cascadingDeleteEH(T t) {
        if (delete(t)) {
            return;
        }
        String e = String.format("tuple of type %s and id=%s does not exist and can't be deleted", TupleTypeName(), t.id);
        throw new RuntimeException(e);
    }

    public void delete() {
        for (T tt : tuples()) {
            delete(tt);
        }
    }

    // General operations that can be performed on all tables
    public boolean contains(T t) { // does table contain tuple t
        for (T tt : tuples()) {
            if (tt.compareTo(t) == 0) {
                return true;
            }
        }
        return false;
    }

    public Stream<T> stream() {
        return tuples().stream();
    }

    public int count() {
        return tuples().size();
    }

    public int size() {  // same as count
        return tuples().size();
    }

    public boolean isEmpty() {
        return count() == 0;
    }

    public boolean isNotEmpty() {
        return count() != 0;
    }

    public abstract void makeAndAdd(Tuple t);

    public abstract TBL New();

    public TBL unique() {
        TBL unique = New();
        for (T t : tuples()) {
            if (!unique.contains(t)) {
                unique.add(t);
            }
        }
        return unique;
    }

    /**
     * form of projection over a column
     * @param getKey  function that returns the key of a T tuple
     * @return table where all key values are unique
     */
    public TBL unique(Function<T, String> getKey) {
        HashMap<String, T> uneek = new HashMap<>();
        TBL result = New();

        for (T t : tuples()) {
            String val = getKey.apply(t);
            uneek.put(val, t);
        }
        for (T j : uneek.values()) {
            result.add(j);
        }
        return result;
    }

    public TBL intersect(TBL x) {
        TBL result = New();
        for (T t : tuples()) {
            if (x.contains(t)) {
                result.add(t);
            }
        }
        return result;
    }

    public TBL union(TBL x) {
        TBL result = New();
        for (T t : tuples()) {
            result.add(t);
        }
        for (T t : x.tuples()) {
            if (!result.contains(t)) {
                result.add(t);
            }
        }
        return result;
    }

    public T first(Predicate<T> p) {
        for (T t : tuples()) {
            if (p.test(t)) {
                return t;
            }
        }
        return null;
    }

    public T getTupleEH(String id) {
        T result = getTuple(id);
        if (result == null) {
            throw new RuntimeException("no tuple with id=" + id + " in table " + TupleTypeName());
        }
        return result;
    }

    public T getTuple(String id) {
        for (T t : tuples()) {
            if (t.id.equals(id)) {
                return t;
            }
        }
        return null;
    }

    public T follow(String id) {
        return getTuple(id);
    }

    public T followEH(String id) {
        return getTupleEH(id);
    }

    public T follow(String id, ErrorReport er) {
        return getTuple(id, er);
    }

    public T getFirst() {
        return tuples().getFirst();
    }

    public T getTuple(String id, ErrorReport er) {
        try {
            return getTupleEH(id);
        } catch (Exception e) {
            er.add("No tuple in table %s has id=%s", TupleTypeName(), id);
            return null;
        }
    }

    public TBL select(Predicate<T> p) {
        TBL shortened = New();
        tuples().stream().filter(p).forEach(t -> shortened.add(t));
        return shortened;
    }

    public TBL filter(Predicate<T> p) { // same as select
        return select(p);
    }

    public TBL reject(Predicate<T> p) {  // same as select (negate predicate)
        return select(p.negate());
    }

    public TBL sort(Comparator<T> comp) {
        LinkedList<T> l = tuples();
        Collections.sort(l, comp);
        TBL newTab = New();
        newTab.tuples = l;
        return newTab;
    }

    public TBL duplicates() {
        TBL dups = New();
        TBL unique = New();
        for (T t : tuples()) {
            if (!unique.contains(t)) {
                unique.add(t);
            } else if (!dups.contains(t)) {
                dups.add(t);
            }
        }
        return dups;
    }

    public void error(ErrorReport er, String format, Function<T, String>... fun) {
        for (T t : tuples()) {
            Object[] sargs = new Object[fun.length];
            for (int i = 0; i < fun.length; i++) {
                sargs[i] = fun[i].apply(t);
            }
            er.add(format, sargs);
        }
    }

    public boolean equals(TBL table) {
        if (table.count() != count()) {
            return false;
        }
        for (T out : this.tuples()) {
            if (!table.contains(out)) {
                return false;
            }
        }
        for (T in : tuples()) {
            if (!contains(in)) {
                return false;
            }
        }
        return true;
    }

    public void forEach(Consumer<T> action) {
        if (count() == 0) {
            return;
        }
        this.stream().forEach(t -> action.accept(t));
    }

    public abstract void mastHead(PrintStream out);

    public boolean anyMatch(Predicate<T> p) {
        return this.stream().anyMatch(p);
    }

    public boolean allMatch(Predicate<T> p) {
        return this.stream().allMatch(p);
    }

    public boolean noMatch(Predicate<T> p) {
        return !this.stream().anyMatch(p);
    }

    public boolean exists(Predicate<T> p) {
        return anyMatch(p);
    }

    public boolean exists() {
        return size() > 0;
    }

    public boolean notExist() {
        return size() == 0;
    }

    public boolean notExist(Predicate<T> p) {
        return !anyMatch(p);
    }

    // print utilities
    public void print() {
        String fmt = pformat();
        printEngine(System.out, fmt, tuples());
    }

    public void prettyPrint() {
        printEngine(System.out, ppformat(false), tuples());
    }

    public void printLocal(PrintStream out) {
        String fmt = pformat();
        printEngine(out, pformat(), tuplesLocal());
    }

    public void prettyprintLocal(PrintStream out) {
        String fmt = ppformat(true);
        printEngine(out, fmt, tuplesLocal());
    }

    public void print(PrintStream out) {
        printEngine(out, pformat(), tuples());
    }

    public void prettyprint(PrintStream out) {
        printEngine(out, ppformat(false), tuples());
    }

    public abstract String pformat();

    public abstract String ppformat(boolean local);

    void printEngine(PrintStream out, String fmt, LinkedList<T> tuplelist) {
        mastHead(out);
        for (T t : tuplelist) {
            String fmts = t.toString(fmt);
            out.print(fmts);
        }
        out.println();
    }

    // Aggregation and groupby utilities
    public void aggregate(OTable tbl, Reducer... agg) {
        // Step 1: reset accumulators and count number of results
        int n = 0;
        for (Reducer r : agg) {
            r.reset();
            n = n + r.resultCount();
        }

        // Step 2: aggregate
        for (T t : tuples()) {
            for (Reducer r : agg) {
                r.add(t);
            }
        }
        // Step 3: create an OTuple from aggregated values
        OTuple a = new OTuple(n);
        int x = 0;  // result index
        for (int i = 0; i < agg.length; i++) {
            for (int j = 0; j < agg[i].resultCount(); j++) {
                a.set(x++, agg[i].getAccumulated(j));
            }
        }

        // Step 4: add tuple to table
        tbl.add(a);
    }

    public Stream<TBL> groupBy(Function<T, String> grouper) {
        // Step 1: partition 'this' tbl by fields[]
        HashMap<String, TBL> gb = new HashMap<>();
        for (T t : tuples()) {
            String key = grouper.apply(t);
            TBL tl = gb.get(key);
            if (tl == null) {
                tl = New();
                gb.put(key, tl);
            }
            tl.add(t);
        }
        return gb.values().stream();
    }

    public void groupBy(Function<T, String> grouper, Consumer<TBL> action) {
        // Step 1: partition 'this' tbl by fields[]
        HashMap<String, TBL> gb = new HashMap<>();
        for (T t : tuples()) {
            String key = grouper.apply(t);
            TBL tl = gb.get(key);
            if (tl == null) {
                tl = New();
                gb.put(key, tl);
            }
            tl.add(t);
        }

        // Step 2: do a foreach over the grouped tables
        for (TBL pt : gb.values()) {
            action.accept(pt);
        }
    }

    // built-in table constraints
    public void isNotNull(String colName, Function<T, String> check, ErrorReport er) {
        for (T t : tuples()) {
            if (check.apply(t) == null) {
                er.add("%s with id=%s has null %s", TupleTypeName(), t.id, colName);
            }
        }
    }

    // input table to noDupls is a single column table
    public void isUnique(String colName, ErrorReport er) {
        HashMap<String, Boolean> hm = new HashMap<>();
        for (T t : tuples()) {
            String value = t.toString();
            if (hm.containsKey(value)) {
                Boolean b = hm.get(value);
                if (!b) {
                    hm.put(value, false);
                    er.add("%s has duplicate value %s", colName, value);
                }
            }
        }
    }

    public void notExist(Predicate<T> p, String expl, ErrorReport er) {
        for (T t : tuples()) {
            if (p.test(t)) {
                er.add(expl + " %s", t.id);
            }
        }
    }

    public void exists(Predicate<T> p, String expl, ErrorReport er) {
        if (notExist(p)) {
            er.add(expl);
        }
    }

    public void implies(Predicate<T> ifpred, Predicate<T> thenpred, String expl, ErrorReport er) {
        for (T t : tuples()) {
            if (notImplies(ifpred.test(t), thenpred.test(t))) {
                er.add(expl + " %s", t.id);
            }
        }
    }

    private boolean implies(boolean a, boolean b) {
        return !a || b;
    }

    private boolean notImplies(boolean a, boolean b) {
        return !implies(a, b);
    }

    public void isLegit(Function<T, Object> access, String field, ErrorReport er) {
        tuples.stream().filter(t -> access.apply(t) == null)
                .forEach(x -> er.add("%s(%s..) has %s=null", TupleTypeName(), x.id, field));
    }

}
