package LineParsers;

import M4Parsing.LineParsers.CDLineParsers;
import CD.CDSpec;
import CD.AssocSchema;
import PrologDB.ErrorReport;
import PrologDB.Table;
import PrologDB.TableSchema;
import org.junit.Test;
import static org.junit.Assert.*;

/**
 *
 * @author don
 */
public class AssocLineTest {

    public AssocLineTest() {
    }

    public void doit(String line, String output) {
        CDSpec os = new CDSpec("me");
        TableSchema lts = new TableSchema("left");
        TableSchema rts = new TableSchema("right");
        os.add(lts);
        os.add(rts);
        ErrorReport er = new ErrorReport();
        //parseAssocLine(String line, int lineno, Schema dbs, ErrorReport er)
        CDLineParsers.parseAssocLine(line, 5, os, er);
        if (er.hasErrors()) {
            fail(er.toString());
        }
        for (AssocSchema as : os.getAssocSchemas()) {
            String s  = as.toString().trim();
            if (!output.trim().equals(s)) {
                fail(output.trim() +"!="+ s);
            }
        }
    }

    public void doitErr(String line, String msg) {
        CDSpec os = new CDSpec("me");
        ErrorReport er = new ErrorReport();
        CDLineParsers.parseAssocLine(line, 5, os, er);
        String e= er.toString().trim();
        if (e.equals(msg))
            return;
        System.out.println(e);
        System.out.println(msg);
        fail();
    }

    @Test
    public void t1() {
        doit("assoc(one, left,leftrole, DIAMOND, right, rightrole,BLACK_DIAMOND ).", 
                "assoc(one,left,leftrole,DIAMOND,right,rightrole,BLACK_DIAMOND).");
    }

    @Test
    public void t2() {
        doit("assoc   (abc , left,leftrole, BLACK_DIAMOND, right, rightrole,DIAMOND ).", 
                "assoc(abc,left,leftrole,BLACK_DIAMOND,right,rightrole,DIAMOND).");
    }
    
    @Test
    public void t3() {
        doit("assoc   (x, left,leftrole, BLANK, right, rightrole,BLANK).", 
                "assoc(x,left,leftrole,BLANK,right,rightrole,BLANK).");
    }

    @Test
    public void b1() {
        doitErr("assoc(a,b,c,d,e.f).", 
                "line 5: parse error 'comma' expected at >>.f).");
    }

    @Test
    public void b2() {
        doitErr("assoc(a,b,c,d)", 
                "line 5: parse error 'comma' expected at >>)");
    }
    
    @Test
    public void T1() {
        doit("assoc one  left leftrole BLANK -- right  rightrole BLANK.", 
                "assoc(one,left,leftrole,BLANK,right,rightrole,BLANK).");
    }
    
    @Test
    public void T2e() {
        doitErr("assoc   abc   left  leftrole  BLACK_DIAMOND -- right  rightrole DIAMOND ) .", 
                "line 5: parse error 'dot' expected at >>) .");
    }
    
    @Test
    public void T2() {
        doit("assoc   abc   left  leftrole  BLACK_DIAMOND -- right  rightrole DIAMOND.", 
                "assoc(abc,left,leftrole,BLACK_DIAMOND,right,rightrole,DIAMOND).");
    }
    
    @Test
    public void T3() {
        doit("assoc   x left leftrole BLANK -- right  rightrole BLANK .", 
                "assoc(x,left,leftrole,BLANK,right,rightrole,BLANK).");
    }
    
    @Test
    public void T4() {
        doit("assoc   x left leftrole ONE -- right  rightrole STAR .", 
                "assoc(x,left,leftrole,BLACK_DIAMOND,right,rightrole,BLANK).");
    }
    
    @Test
    public void T5() {
        doit("assoc   x left leftrole ONE -- right  rightrole ONESTAR .", 
                "assoc(x,left,leftrole,BLACK_DIAMOND,right,rightrole,ONESTAR).");
    }
    
    @Test
    public void T6() {
        doit("assoc   x left leftrole OPT -- right  rightrole ONESTAR .", 
                "assoc(x,left,leftrole,DIAMOND,right,rightrole,ONESTAR).");
    }
    
    @Test
    public void T7() {
        doit("assoc   x left leftrole 3:5 -- right  rightrole 8:9 .", 
                "assoc(x,left,leftrole,3:5,right,rightrole,8:9).");
    }
}
