package Aocl.Alleg;

import PrologDB.DB;
import PrologDB.Table;
import PrologDB.ErrorReport;
import java.io.PrintStream;

public class Database {

    public EmpTable Emp = EmpTable.BaseRelation(this);
    public DeptTable Dept = DeptTable.BaseRelation(this);
    public DivTable Div = DivTable.BaseRelation(this);
    public AncTable Anc = AncTable.BaseRelation(this);
    public JobTable Job = JobTable.BaseRelation(this);

    private DB db;

    public Database() {}

    public Database(String filename) {
        // Step 1: Read the database, throw exception if errors found and halt
        db = DB.read(filename);

        // Step 2: access MDELite tables
        Table _Emp = db.getTableEH("Emp");
        Table _Dept = db.getTableEH("Dept");
        Table _Div = db.getTableEH("Div");
        Table _Anc = db.getTableEH("Anc");
        Table _Job = db.getTableEH("Job");

        // Step 3: pass1 -- load primitive data of all non-abstract + non-virtual tables
        _Emp.tuplesLocal().forEach(t->Emp.makeAndAdd(t));
        _Dept.tuplesLocal().forEach(t->Dept.makeAndAdd(t));
        _Div.tuplesLocal().forEach(t->Div.makeAndAdd(t));
        _Anc.tuplesLocal().forEach(t->Anc.makeAndAdd(t));
        _Job.tuplesLocal().forEach(t->Job.makeAndAdd(t));

        // Step 4: pass2 -- now replace MDELite tuple ids with Java object ids
        ErrorReport er = new ErrorReport();
        Emp.tuplesLocal().forEach(t->t.finishTuple(this,er));
        Dept.tuplesLocal().forEach(t->t.finishTuple(this,er));
        Div.tuplesLocal().forEach(t->t.finishTuple(this,er));
        Anc.tuplesLocal().forEach(t->t.finishTuple(this,er));
        Job.tuplesLocal().forEach(t->t.finishTuple(this,er));

        // Step 5: pass3 -- EXPERIMENTAL init of virtual tables --- must manually fill in expressions

        er.printEH(System.out);
    }

    public void print() { print(System.out); }

    public void prettyprint() { prettyprint(System.out); }

    public void prettyprint(PrintStream out) {
        out.format("dbase(Alleg,[Emp,Dept,Div,Anc,Job]).\n\n");
        Emp.prettyprintLocal(out);
        Dept.prettyprintLocal(out);
        Div.prettyprintLocal(out);
        Anc.prettyprintLocal(out);
        Job.prettyprintLocal(out);
    }

    public void print(PrintStream out) {
        out.format("dbase(Alleg,[Emp,Dept,Div,Anc,Job]).\n\n");
        Emp.printLocal(out);
        Dept.printLocal(out);
        Div.printLocal(out);
        Anc.printLocal(out);
        Job.printLocal(out);
    }

}
