package Aocl.Star;

import PrologDB.DB;
import PrologDB.Table;
import PrologDB.ErrorReport;
import java.io.PrintStream;

public class Database {

    public XTable X = XTable.BaseRelation(this);
    public YTable Y = YTable.BaseRelation(this);
    public ZTable Z = ZTable.BaseRelation(this);
    public WTable W = WTable.BaseRelation(this);
    public CTable C = CTable.BaseRelation(this);
    public xcTable xc = xcTable.BaseRelation(this);
    public ycTable yc = ycTable.BaseRelation(this);
    public zcTable zc = zcTable.BaseRelation(this);
    public wcTable wc = wcTable.BaseRelation(this);

    private DB db;

    public Database() {}

    public Database(String filename) {
        // Step 1: Read the database, throw exception if errors found and halt
        db = DB.read(filename);

        // Step 2: access MDELite tables
        Table _X = db.getTableEH("X");
        Table _Y = db.getTableEH("Y");
        Table _Z = db.getTableEH("Z");
        Table _W = db.getTableEH("W");
        Table _C = db.getTableEH("C");
        Table _xc = db.getTableEH("xc");
        Table _yc = db.getTableEH("yc");
        Table _zc = db.getTableEH("zc");
        Table _wc = db.getTableEH("wc");

        // Step 3: pass1 -- load primitive data of all non-abstract + non-virtual tables
        _X.tuplesLocal().forEach(t->X.makeAndAdd(t));
        _Y.tuplesLocal().forEach(t->Y.makeAndAdd(t));
        _Z.tuplesLocal().forEach(t->Z.makeAndAdd(t));
        _W.tuplesLocal().forEach(t->W.makeAndAdd(t));
        _C.tuplesLocal().forEach(t->C.makeAndAdd(t));
        _xc.tuplesLocal().forEach(t->xc.makeAndAdd(t));
        _yc.tuplesLocal().forEach(t->yc.makeAndAdd(t));
        _zc.tuplesLocal().forEach(t->zc.makeAndAdd(t));
        _wc.tuplesLocal().forEach(t->wc.makeAndAdd(t));

        // Step 4: pass2 -- now replace MDELite tuple ids with Java object ids
        ErrorReport er = new ErrorReport();
        X.tuplesLocal().forEach(t->t.finishTuple(this,er));
        Y.tuplesLocal().forEach(t->t.finishTuple(this,er));
        Z.tuplesLocal().forEach(t->t.finishTuple(this,er));
        W.tuplesLocal().forEach(t->t.finishTuple(this,er));
        C.tuplesLocal().forEach(t->t.finishTuple(this,er));
        xc.tuplesLocal().forEach(t->t.finishTuple(this,er));
        yc.tuplesLocal().forEach(t->t.finishTuple(this,er));
        zc.tuplesLocal().forEach(t->t.finishTuple(this,er));
        wc.tuplesLocal().forEach(t->t.finishTuple(this,er));

        // Step 5: pass3 -- EXPERIMENTAL init of virtual tables --- must manually fill in expressions

        er.printEH(System.out);
    }

    public void print() { print(System.out); }

    public void prettyprint() { prettyprint(System.out); }

    public void prettyprint(PrintStream out) {
        out.format("dbase(Star,[X,Y,Z,W,C,xc,yc,zc,wc]).\n\n");
        X.prettyprintLocal(out);
        Y.prettyprintLocal(out);
        Z.prettyprintLocal(out);
        W.prettyprintLocal(out);
        C.prettyprintLocal(out);
        xc.prettyprintLocal(out);
        yc.prettyprintLocal(out);
        zc.prettyprintLocal(out);
        wc.prettyprintLocal(out);
    }

    public void print(PrintStream out) {
        out.format("dbase(Star,[X,Y,Z,W,C,xc,yc,zc,wc]).\n\n");
        X.printLocal(out);
        Y.printLocal(out);
        Z.printLocal(out);
        W.printLocal(out);
        C.printLocal(out);
        xc.printLocal(out);
        yc.printLocal(out);
        zc.printLocal(out);
        wc.printLocal(out);
    }

}
