package Aocl.startrek;

import PrologDB.DB;
import PrologDB.Table;
import PrologDB.ErrorReport;
import java.io.PrintStream;

public class Database {

    public crewmanTable crewman = crewmanTable.BaseRelation(this);
    public officerTable officer = officerTable.BaseRelation(this);
    public medicalTable medical = medicalTable.BaseRelation(this);
    public postTable post = postTable.BaseRelation(this);
    public asgTable asg = asgTable.BaseRelation(this);

    private DB db;

    public Database() {}

    public Database(String filename) {
        // Step 1: Read the database, throw exception if errors found and halt
        db = DB.read(filename);

        // Step 2: access MDELite tables
        Table _crewman = db.getTableEH("crewman");
        Table _officer = db.getTableEH("officer");
        Table _medical = db.getTableEH("medical");
        Table _post = db.getTableEH("post");
        Table _asg = db.getTableEH("asg");

        // Step 3: pass1 -- load primitive data of all non-abstract + non-virtual tables
        _crewman.tuplesLocal().forEach(t->crewman.makeAndAdd(t));
        _officer.tuplesLocal().forEach(t->officer.makeAndAdd(t));
        _medical.tuplesLocal().forEach(t->medical.makeAndAdd(t));
        _post.tuplesLocal().forEach(t->post.makeAndAdd(t));
        _asg.tuplesLocal().forEach(t->asg.makeAndAdd(t));

        // Step 4: pass2 -- now replace MDELite tuple ids with Java object ids
        ErrorReport er = new ErrorReport();
        crewman.tuplesLocal().forEach(t->t.finishTuple(this,er));
        officer.tuplesLocal().forEach(t->t.finishTuple(this,er));
        medical.tuplesLocal().forEach(t->t.finishTuple(this,er));
        post.tuplesLocal().forEach(t->t.finishTuple(this,er));
        asg.tuplesLocal().forEach(t->t.finishTuple(this,er));

        // Step 5: pass3 -- EXPERIMENTAL init of virtual tables --- must manually fill in expressions

        er.printEH(System.out);
    }

    public void print() { print(System.out); }

    public void prettyprint() { prettyprint(System.out); }

    public void prettyprint(PrintStream out) {
        out.format("dbase(startrek,[crewman,officer,medical,post,asg]).\n\n");
        crewman.prettyprintLocal(out);
        officer.prettyprintLocal(out);
        medical.prettyprintLocal(out);
        post.prettyprintLocal(out);
        asg.prettyprintLocal(out);
        out.format("subtable(crewman,[officer,medical]).\n");
    }

    public void print(PrintStream out) {
        out.format("dbase(startrek,[crewman,officer,medical,post,asg]).\n\n");
        crewman.printLocal(out);
        officer.printLocal(out);
        medical.printLocal(out);
        post.printLocal(out);
        asg.printLocal(out);
        out.format("subtable(crewman,[officer,medical]).\n");
    }

}
